#include "catalog.h"
#include "match.h"
#include "align.h"
#include "image_data.h"
#include "image_io.h"
#include <stdio.h>

void
DumpTriangles3D(const std::vector<s3d_merge::Point3D> &points,
                const std::vector<s3d_merge::Triangle> &triangles)
{
    for(size_t i=0;i<triangles.size();i++)
    {
        printf("%g, %g, %g\n", 
               points[triangles[i].m_points[0]][0],
               points[triangles[i].m_points[0]][1],
               points[triangles[i].m_points[0]][2]);

        printf("%g, %g, %g\n", 
               points[triangles[i].m_points[1]][0],
               points[triangles[i].m_points[1]][1],
               points[triangles[i].m_points[1]][2]);

        printf("%g, %g, %g\n", 
               points[triangles[i].m_points[2]][0],
               points[triangles[i].m_points[2]][1],
               points[triangles[i].m_points[2]][2]);

        printf("%g, %g, %g\n\n\n", 
               points[triangles[i].m_points[0]][0],
               points[triangles[i].m_points[0]][1],
               points[triangles[i].m_points[0]][2]);
    }
}


class CandidateMatch
{
public:
    CandidateMatch(size_t from, size_t to)
        : m_from(from)
        , m_to(to)
        , m_score(0)
    {
    }


    size_t From(void) const
    {
        return m_from;
    }

    size_t To(void) const
    {
        return m_to;
    }

    size_t Score(void) const
    {
        return m_score;
    }

    void IncrementScore(void)
    {
        m_score++;
    }

private:
    size_t m_from,
           m_to, 
           m_score;
};

// Sort into reverse order of score so we can cull from the back end of a vector.
class CandidateMatchComparator : public std::binary_function<const CandidateMatch &, const CandidateMatch &, bool>
{
public:
    bool operator()(const CandidateMatch &a, const CandidateMatch &b)
    {
        return a.Score() > b.Score();
    }
};

void
DumpTriangles2D(const std::vector<s3d_merge::Point2D> &points,
                const std::vector<s3d_merge::Triangle> &triangles)
{
    for(size_t i=0;i<triangles.size();i++)
    {
        printf("%g, %g\n", 
               points[triangles[i].m_points[0]][0],
               points[triangles[i].m_points[0]][1]);

        printf("%g, %g\n", 
               points[triangles[i].m_points[1]][0],
               points[triangles[i].m_points[1]][1]);

        printf("%g, %g\n", 
               points[triangles[i].m_points[2]][0],
               points[triangles[i].m_points[2]][1]);

        printf("%g, %g\n\n\n", 
               points[triangles[i].m_points[0]][0],
               points[triangles[i].m_points[0]][1]);
    }
}

bool 
CheckCandidateDistanceLengthRatio(
    const std::vector<s3d_merge::Point2D> &img_points,
    const std::vector<s3d_merge::Triangle> &img_triangles,
    const std::vector<s3d_merge::Point2D> &img_centroids,
    const std::vector<s3d_merge::Point3D> &ref_points,
    const std::vector<s3d_merge::Triangle> &ref_triangles,
    const std::vector<s3d_merge::Point3D> &ref_centroids,
    const CandidateMatch &a,
    const CandidateMatch &b)

{
    return false;
}

bool
CheckCandidateLongSideOrientations(
    const std::vector<s3d_merge::Point2D> &img_points,
    const std::vector<s3d_merge::Triangle> &img_triangles,
    const std::vector<s3d_merge::Point2D> &img_centroids,
    const std::vector<s3d_merge::Point3D> &ref_points,
    const std::vector<s3d_merge::Triangle> &ref_triangles,
    const std::vector<s3d_merge::Point3D> &ref_centroids,
    const CandidateMatch &a,
    const CandidateMatch &b)
{
    
    return false;
}

bool
CheckCandidateRelativeScales(
    const std::vector<s3d_merge::Point2D> &img_points,
    const std::vector<s3d_merge::Triangle> &img_triangles,
    const std::vector<s3d_merge::Point2D> &img_centroids,
    const std::vector<s3d_merge::Point3D> &ref_points,
    const std::vector<s3d_merge::Triangle> &ref_triangles,
    const std::vector<s3d_merge::Point3D> &ref_centroids,
    const CandidateMatch &a,
    const CandidateMatch &b)
{

    return false;
}



int main(int argc, char ** argv)
{
    std::vector<s3d_merge::CatalogueRef> ref_stars;

    double cutoff(9);

    {
        std::vector<s3d_merge::CatalogueRef> all_stars;
        printf("Reading catalogue\n");
        s3d_merge::ParseYaleCatalogue("bsc/bsc5.dat", all_stars);

        printf("Retrieved records for %li stars\n", all_stars.size());
        s3d_merge::MagnitudeComparator cmp;

        std::sort(all_stars.begin(), all_stars.end(), cmp);

        for(size_t i=0;i<all_stars.size();i++)
        {
            if(all_stars[i].GetMagnitude() < cutoff)
            {
                ref_stars.push_back(all_stars[i]);
            }
            else
            {
                break;
            }
        }
    }

    printf("Retained %li stars with magnitude < %lf\n", ref_stars.size(), cutoff);

    std::vector<s3d_merge::Point3D> ref_points;
    std::vector<s3d_merge::Point2D> cur_points;
    ref_points.reserve(ref_stars.size());

    printf("building vector tree\n");

    for(size_t i=0;i<ref_stars.size();i++)
    {
        double tmp[3];
        ref_stars[i].Get3DVector(tmp);
        ref_points.push_back(s3d_merge::Point3D(tmp));
    }

    s3d_merge::KDTree3D ref_tree(ref_points.begin(), ref_points.end());

    printf("built vector tree\n");

    std::vector<s3d_merge::Triangle> ref_triangles;

    printf("building triangles\n");


    {
        double min_scale(0.5);
        double max_scale(0.75);

        while(max_scale < 30)
        {
            BuildTriangles3D(ref_tree, ref_points,  min_scale,  max_scale, 1, ref_triangles);
            min_scale*=1.25;
            max_scale*=1.25;
        }
    }

    // Remove duplicates
    if(ref_triangles.size())

    {
        std::sort(ref_triangles.begin(),ref_triangles.end());
        std::vector<s3d_merge::Triangle>::iterator it = std::unique(ref_triangles.begin(),ref_triangles.end());
        if(it != ref_triangles.end())
        {
            ref_triangles.erase(it, ref_triangles.end());
        }
        // Free space.
        std::vector<s3d_merge::Triangle> tmp_triangles;
        tmp_triangles.reserve(ref_triangles.size());
        tmp_triangles.insert(tmp_triangles.end(), ref_triangles.begin(), ref_triangles.end());
        tmp_triangles.swap(ref_triangles);
    }
    //DumpTriangles3D(ref_points, ref_triangles);

    std::vector<s3d_merge::Point2D> ref_shapes;
    std::vector<s3d_merge::Point3D> ref_centroids;

    ref_shapes.reserve(ref_triangles.size());
    ref_centroids.reserve(ref_triangles.size());

    BuildTriangleStats3D(ref_triangles, ref_points, ref_shapes, ref_centroids);

    printf("building triangle tree\n");
    s3d_merge::KDTree2D ref_shapes_tree(ref_shapes.begin(), ref_shapes.end());
    printf("built triangle tree with %li elements\n", ref_shapes.size());

    if(argc > 1)
    {
        s3d_merge::ImageData *id;

        id = s3d_merge::LoadImageData(argv[1]);

        if(id && id->Image())
        {
            std::vector<s3d_merge::StarRef> stars;
            s3d_merge::FindStars(id->Image(), 30, 30, 15, 10, 400, stars);
            std::vector<s3d_merge::Point2D> img_points;
            std::vector<s3d_merge::Point2D> img_orig_points;

            {
                double scale; 
                s3d_merge::Point2D translate;
                translate[0] = id->Image()->get_width();
                translate[1] =  id->Image()->get_height();
                translate *= 0.5;

                if(id->Image()->get_width() > id->Image()->get_height())
                {
                    scale = translate[0];
                }
                else
                {
                    scale = translate[1];
                }

                for(size_t i=0;i<stars.size();i++)
                {
                    s3d_merge::Point2D tmp;
                    tmp = ((stars[i].GetPoint() - translate) / scale);
                    tmp[1] = -tmp[1]; // coordinate hack for fitting compatibility.
                    img_points.push_back(tmp);
                    img_orig_points.push_back(stars[i].GetPoint());
                }
            }

            s3d_merge::KDTree2D img_tree(img_points.begin(), img_points.end());

            std::vector<s3d_merge::Triangle> img_triangles;
            std::vector<s3d_merge::Point2D> img_shapes;
            std::vector<s3d_merge::Point2D> img_centroids;

            {
                double min_scale(0.001);
                double max_scale(0.0015);
        
                while(max_scale < 1)
                {
                    BuildTriangles2D(img_tree, img_points,  min_scale,  max_scale, 1, img_triangles);
                    min_scale*=1.25;
                    max_scale*=1.25;
                }
            }
            {
                std::sort(img_triangles.begin(),img_triangles.end());
                std::vector<s3d_merge::Triangle>::iterator it = std::unique(img_triangles.begin(),img_triangles.end());
                if(it != img_triangles.end())
                {
                    img_triangles.erase(it, img_triangles.end());
                }
                // Free space.
                std::vector<s3d_merge::Triangle> tmp_triangles;
                tmp_triangles.reserve(img_triangles.size());
                tmp_triangles.insert(tmp_triangles.end(), img_triangles.begin(), img_triangles.end());
                tmp_triangles.swap(img_triangles);
            }

            printf("built %li triangles\n", img_triangles.size());
            //DumpTriangles2D(img_points, img_triangles);

            BuildTriangleStats2D(img_triangles, 
                                 img_points,
                                 img_shapes,
                                 img_centroids);

            // (score(from, to))
            std::vector<CandidateMatch> candidate_matches;

            candidate_matches.reserve(img_shapes.size() * 20);
            for(size_t i=0;i<img_shapes.size();i++)
            {
                std::vector<size_t> found;
                found.clear();
                ref_shapes_tree.FindClosestItems(img_shapes[i], 10, found);

                for(size_t j=0;j<found.size();j++)
                {
                    candidate_matches.push_back(CandidateMatch(i, found[j]));
                }
            }
            printf("found %li candidate matches.\n", candidate_matches.size());

            for(size_t i=0;i<5;i++) // Outer loop.
            {
                for(size_t j=0;j<500;j++) // Iterations between culls.
                {
                    for(size_t k=0;k<candidate_matches.size();k++)
                    {
                        size_t other;
                        bool pass;

                        do{
                            other = size_t(rand()) % candidate_matches.size();
                        } while((candidate_matches[k].From() != candidate_matches[other].From()) &&
                                (candidate_matches[k].From() != candidate_matches[other].To()));

                        // We perform a simple test to see whether these two matches are consistent with one another
                        // on the input and output side.  If they are then the scores of both go up.  If not then they
                        // stay the same.
                        pass = false;
                        switch(j%3)
                        {
                        case 0:
                            // Ratio of ong side to distance between centroids
                            pass = CheckCandidateDistanceLengthRatio(img_points, img_triangles, img_centroids,
                                                                     ref_points, ref_triangles, ref_centroids,
                                                                     candidate_matches[k],
                                                                     candidate_matches[other]);
                            break;
                        case 1:
                            // Dot product of long side directions
                            pass = CheckCandidateLongSideOrientations(img_points, img_triangles, img_centroids,
                                                                      ref_points, ref_triangles, ref_centroids,
                                                                      candidate_matches[k],
                                                                      candidate_matches[other]);
                            break;
                        case 2:
                            // Relative sizes of triangles.
                            pass = CheckCandidateRelativeScales(img_points, img_triangles, img_centroids,
                                                                ref_points, ref_triangles, ref_centroids,
                                                                candidate_matches[k],
                                                                candidate_matches[other]);

                        }

                        // Test passed;
                        if(pass)
                        {
                            candidate_matches[k].IncrementScore();
                            candidate_matches[other].IncrementScore();
                        }
                    }
                }
                // Throw away worst half of the results.
                {
                    CandidateMatchComparator cmp;
                    std::sort(candidate_matches.begin(), 
                              candidate_matches.end(), 
                              cmp);
                    
                    if(candidate_matches.size() > 16)
                    {
                        std::vector<CandidateMatch>::iterator new_end =
                            candidate_matches.begin() + (candidate_matches.size() / 2);
                        candidate_matches.erase(new_end, candidate_matches.end());
                        
                    }
                }
            }
        }
    }
    return 0;
}
