#ifndef __PREFERENCES_H__
#define __PREFERENCES_H__

#include <stdint.h>
#include <cstdlib>

namespace s3d_merge
{

class Preferences
{
public:
    Preferences();
    Preferences(const char *);

    void SetDefaults(void);
    void LoadPreferences(const char *);

    void DumpPreferences(void);

    // Star finder
    uint8_t GetFindStarsTriggerThreshold(void) const
    {
        return m_findStarsTriggerThreshold;
    }

    uint8_t GetFindStarsExitThreshold(void) const
    {
        return m_findStarsExitThreshold;
    }

    size_t GetFindStarsMaxPixels(void) const
    {
        return m_findStarsMaxPixels;
    }

    size_t GetFindStarsMinPixels(void) const
    {
        return m_findStarsMinPixels;
    }

    size_t GetFindStarsMaxStars(void) const
    {
        return m_findStarsMaxStars;
    }

    uint32_t GetFindStarsBlurRadius(void) const
    {
        return m_findStarsBlurRadius;
    }

    // Image align
    double GetAlignStarsConsensusError(void) const
    {
        return m_alignStarsConsensusError;
    }

    // Image remap
    uint32_t GetRemapImageAAType(void) const
    {
        return m_remapImageAAType;
    }

private:
    bool ParseField(const char *, const char *);

    uint8_t m_findStarsTriggerThreshold;
    uint8_t m_findStarsExitThreshold;
    size_t  m_findStarsMinPixels;
    size_t  m_findStarsMaxPixels;
    size_t  m_findStarsMaxStars;
    uint32_t m_findStarsBlurRadius;

    double m_alignStarsConsensusError;

    uint32_t m_remapImageAAType;
};

};

#endif
